from pytracking.tracker.base import BaseTracker
import torch
import torch.nn.functional as F
import math
import time
from pytracking import dcf, TensorList
from pytracking.features.preprocessing import numpy_to_torch
from pytracking.utils.plotting import show_tensor, plot_graph
from pytracking.features.preprocessing import sample_patch_multiscale, sample_patch_transformed
from pytracking.features import augmentation
import ltr.data.bounding_box_utils as bbutils
from ltr.models.target_classifier.initializer import FilterInitializerZero
from ltr.models.layers import activation

from pytracking.tracker.STM.stm_models import STM
import numpy as np
from pytracking.utils import processing_utils as prutils
import cv2
from pytracking.mask_to_disk import save_mask
from pytracking.bbox_fit import fit_bbox_to_mask
import copy

import seaborn as sns
import matplotlib.pyplot as plt


class STM_tracker(BaseTracker):

    multiobj_mode = 'parallel'
    def poly_to_aabbox(self, x_, y_):
        # keep the center and area of the polygon
        # change aspect ratio of the original bbox
        cx = np.mean(x_)
        cy = np.mean(y_)
        x1 = np.min(x_)
        x2 = np.max(x_)
        y1 = np.min(y_)
        y2 = np.max(y_)
        A1 = np.linalg.norm(np.array([x_[0], y_[0]]) - np.array([x_[1], y_[1]])) * \
             np.linalg.norm(np.array([x_[1], y_[1]]) - np.array([x_[2], y_[2]]))
        A2 = (x2 - x1) * (y2 - y1)
        s = np.sqrt(A1 / A2)
        w = s * (x2 - x1) + 1
        h = s * (y2 - y1) + 1
        return np.array([cx - w / 2, cy - h / 2, w, h])

    def initialize_model(self, cuda_device=None):
        self.AMB_model = STM()
        if torch.cuda.is_available():
            self.AMB_model.cuda()
        self.AMB_model.eval()  # turn-off BN
        for p in self.AMB_model.parameters():
            p.requires_grad = False

        pth_path = '/home/v-xfei/Public/project/pytracking/pytracking/networks/STM_weights.pth'
        # print("###############" + pth_path + "###############")
        self.AMB_model.load_param(torch.load(pth_path))


    def initialize(self, image, state, init_mask=None):
        # Initialize some stuff
        self.frame_num = 1

        # Initialize network
        self.initialize_model()

        self.init_segmentation(image, state, init_mask=init_mask)


    def init_segmentation(self, image, bb, init_mask=None):
        self.seg_frame_num = 1
        self.segm_search_area_factor=1.5
        self.segm_output_sz=256#384
        init_patch_crop, f_ = prutils.sample_target(image, np.array(bb), self.segm_search_area_factor,
                                                    output_sz=self.segm_output_sz)
        """
        init_patch_crop = np.array[384, 384, 3], f_= 1.5609
        """
        self.segmentation_task = True
        if init_mask is not None:
            mask = copy.deepcopy(init_mask).astype(np.float32)
            self.segmentation_task = True
            # self.params.segm_optimize_polygon = True
            # segmentation videos are shorter - therefore larger scale change factor can be used
            self.min_scale_change_factor = 0.9
            self.max_scale_change_factor = 1.1
            self.segm_mask_thr = 0.2
        else:
            """
            mask = np.array[352, 624] 以boundingbox 为mask
            """
            mask = np.zeros((image.shape[0], image.shape[1]), dtype=np.int32)
            if hasattr(self, 'gt_poly'):
                p1 = self.gt_poly[:2]
                p2 = self.gt_poly[2:4]
                p3 = self.gt_poly[4:6]
                p4 = self.gt_poly[6:]
                cv2.fillConvexPoly(mask, np.array([p1, p2, p3, p4], dtype=np.int32), 1)
                mask = mask.astype(np.float32)
            else:
                p1 = bb[:2]
                p2 = [bb[0] + bb[2], bb[1]]
                p3 = [bb[0] + bb[2], bb[1] + bb[3]]
                p4 = [bb[0], bb[1] + bb[3]]
                cv2.fillConvexPoly(mask, np.array([p1, p2, p3, p4], dtype=np.int32), 1)
                mask = mask.astype(np.float32)

        init_mask_patch_np, patch_factor_init = prutils.sample_target(mask, np.array(bb),
                                                                      self.segm_search_area_factor,
                                                                      output_sz=self.segm_output_sz, pad_val=0)

        ##############################################################################################
        self.segm_normalize_mean =  [0.485, 0.456, 0.406]
        self.segm_normalize_std = [0.229, 0.224, 0.225]

        self.segm_normalize_mean = np.array(self.segm_normalize_mean).reshape((1, 1, 3))
        self.segm_normalize_std = np.array(self.segm_normalize_std).reshape((1, 1, 3))

        # normalize input image
        init_patch_norm_ = init_patch_crop.astype(np.float32) / float(255)
        init_patch_norm_ -= self.segm_normalize_mean
        init_patch_norm_ /= self.segm_normalize_std

        # create distance map for discriminative segmentation
        # put image patch and mask to GPU
        init_patch = torch.Tensor(init_patch_norm_)
        init_mask_patch = torch.Tensor(init_mask_patch_np)

        init_mask_patch_AMB_bg_np = 1 - init_mask_patch_np
        init_mask_patch_bg = 1 - init_mask_patch

        self.use_gpu = True

        if self.use_gpu:
            init_patch = init_patch.cuda()
            init_mask_patch = init_mask_patch.cuda()
            init_mask_patch_bg = init_mask_patch_bg.cuda()

        # reshape image for the feature extractor
        init_patch = torch.unsqueeze(init_patch, dim=0).permute(0, 3, 1, 2)  # tensor[1, 3, 384, 384]
        init_mask_patch = torch.unsqueeze(torch.unsqueeze(init_mask_patch, dim=0), dim=0)  # tensor[1, 1, 384, 384]
        init_mask_patch_bg = torch.unsqueeze(torch.unsqueeze(init_mask_patch_bg, dim=0),
                                             dim=0)  # tensor[1, 1, 384, 384]

        init_mask_patch_k = torch.cat([init_mask_patch_bg, init_mask_patch], 1).contiguous()  # [1, k, 384, 384]

        """
         init_patch = tensor[1, 3, 384, 384]
         init_mask_patch = tensor[1, 1, 384, 384]
        """
        if init_mask is None:
            iters = 0
            while iters < 1:
                # Obtain segmentation prediction
                ###################################################################
                with torch.no_grad():  # m_input init_mask_patch_input_stm
                    prev_key, prev_value, _ = self.AMB_model.memorize(init_patch, init_mask_patch_k)
                    # segment
                    logit, ps = self.AMB_model.segment(init_patch, prev_key, prev_value)

                    out = torch.softmax(logit, dim=1)  # tensor [1, k = 6, set_h, set_w]
                    mask_AMB = out[0, 1, :, :].cpu().numpy()
                    mask = mask_AMB.astype(np.float32)
                   # sns.heatmap(mask)
                   # plt.show()
                ###################################################################

                if hasattr(self, 'gt_poly'):
                    # dilate polygon-based mask
                    # dilate only if given mask is made from polygon, not from axis-aligned bb (since rotated bb is much tighter)
                    dil_kernel_sz = max(5, int(round(0.05 * min(self.target_sz).item() * f_)))
                    kernel = np.ones((dil_kernel_sz, dil_kernel_sz), np.uint8)
                    mask_dil = cv2.dilate(init_mask_patch_np, kernel, iterations=1)
                    """
                    膨胀操作
                    """
                    mask = mask * mask_dil
                    """
                    dil_kernel_sz = 5
                    kernel = [5, 5]: 全是1 
                    mask_dil = np.array[384, 384]
                    """
                else:
                    mask = mask * init_mask_patch_np

                target_pixels = np.sum((mask > 0.5).astype(np.float32))
                self.segm_init_target_pixels = target_pixels

                segm_crop_sz = math.ceil(math.sqrt(bb[2] * bb[3]) * self.params.segm_search_area_factor)
                self.vos_mask_result = save_mask(None, mask, segm_crop_sz, bb, image.shape[1], image.shape[0])

                mask_gpu = torch.unsqueeze(torch.unsqueeze(torch.tensor(mask), dim=0), dim=0).cuda()
                iters += 1
        else:
            init_mask_patch_np = (init_mask_patch_np > 0.1).astype(np.float32)
            target_pixels = np.sum((init_mask_patch_np).astype(np.float32))
            self.segm_init_target_pixels = target_pixels
            mask_gpu = torch.unsqueeze(torch.unsqueeze(torch.tensor(init_mask_patch_np), dim=0), dim=0).cuda()



        self.mask_pixels = np.array([np.sum(mask)])  # np.array[1]: 7539.37

        self.pre_frame_input = init_patch
        self.pre_mask_input = torch.cat((1 - mask_gpu, mask_gpu), 1)  # [1, k, 384, 384]
        self.keys = []
        self.values = []

        a = 1

    def segment_target(self, image, pos, sz):
        # pos and sz are in the image coordinates
        # construct new bounding box first
        self.seg_frame_num = self.seg_frame_num + 1

        tlx_ = pos[1] - sz[1] / 2
        tly_ = pos[0] - sz[0] / 2
        w_ = sz[1]
        h_ = sz[0]
        bb = [tlx_.item(), tly_.item(), w_.item(), h_.item()]

        # extract patch
        patch, f_ = prutils.sample_target(image, np.array(bb), self.segm_search_area_factor,
                                          output_sz=self.segm_output_sz)

        segm_crop_sz = math.ceil(math.sqrt(bb[2] * bb[3]) * self.segm_search_area_factor)

        # normalize input image
        init_patch_norm_ = patch.astype(np.float32) / float(255)
        init_patch_norm_ -= self.segm_normalize_mean
        init_patch_norm_ /= self.segm_normalize_std

        # put image patch and mask to GPU
        patch_gpu = torch.Tensor(init_patch_norm_)
        if self.use_gpu:
            patch_gpu = patch_gpu.cuda()
            # reshape image for the feature extractor
            patch_gpu = torch.unsqueeze(patch_gpu, dim=0).permute(0, 3, 1, 2)

        with torch.no_grad():  # m_input init_mask_patch_input_stm
            this_key, this_value, _ = self.AMB_model.memorize(self.pre_frame_input, self.pre_mask_input)

            temp_keys = torch.cat(self.keys + [this_key], dim=1)  # + [this_key]
            temp_values = torch.cat(self.values + [this_value], dim=1)  # + [this_value]

            # segment
            logit, ps = self.AMB_model.segment(patch_gpu, temp_keys, temp_values)

            out = torch.softmax(logit, dim=1)  # tensor [1, k = 6, set_h, set_w]

            mask_AMB = out[0, 1, :, :].cpu().numpy()
            mask = mask_AMB.astype(np.float32)

            mask_gpu = torch.unsqueeze(torch.unsqueeze(torch.tensor(mask), dim=0), dim=0).cuda()  # [1, 1, 384, 384]
            self.pre_frame_input = patch_gpu
            self.pre_mask_input = torch.cat((1 - mask_gpu, mask_gpu), 1).contiguous()  # [1, k, 384, 384]

            if (self.seg_frame_num - 1) % 5 == 0:  # 5, 10
                self.keys.append(this_key)
                self.values.append(this_value)
                pass
            if len(self.keys) > 5:
                #del (self.keys[1])
                #del (self.values[1])
                pass

        mask_real = copy.copy(mask)
        self.segm_mask_thr = 0.5
        mask = (mask > self.segm_mask_thr).astype(np.uint8)

        if cv2.__version__[-5] == '4':
            contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
        else:
            _, contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
        cnt_area = [cv2.contourArea(cnt) for cnt in contours]

        mask = np.zeros(mask.shape, dtype=np.uint8)
        cv2.drawContours(mask, contours, -1, 1, thickness=-1)

        self.vos_mask_result = save_mask(None, mask_real, segm_crop_sz, bb, image.shape[1], image.shape[0])
        # return self.vos_mask_result
        """
                if len(cnt_area) > 0 and len(contours) != 0 and np.max(cnt_area) > 50:  # 1000:
            contour = contours[np.argmax(cnt_area)]  # use max area polygon
            polygon = contour.reshape(-1, 2)

            prbox = np.reshape(cv2.boxPoints(cv2.minAreaRect(polygon)), (4, 2))  # Rotated Rectangle
            prbox_init = copy.deepcopy(prbox)

            prbox_opt = np.array([])

            if self.params.segm_optimize_polygon:
                if not self.segmentation_task:
                    mask = np.zeros(mask.shape, dtype=np.uint8)
                    cv2.drawContours(mask, [contour], -1, 1, thickness=-1)

                    self.track_mask_result = save_mask(mask, mask_real, segm_crop_sz, bb, image.shape[1],
                                                           image.shape[0],
                                                           self.params.masks_save_path)

                t_opt_start_ = time.time()
                prbox_opt_ = fit_bbox_to_mask(mask.astype(np.int32), rotated=self.rotated_bbox)
                bbox_opt_time = time.time() - t_opt_start_
                if prbox_opt_ is not None:
                    A1 = np.linalg.norm(np.array([prbox[0, 0], prbox[0, 1]]) - np.array([prbox[1, 0], prbox[1, 1]])) * \
                         np.linalg.norm(np.array([prbox[1, 0], prbox[1, 1]]) - np.array([prbox[2, 0], prbox[2, 1]]))
                    A_new = np.linalg.norm(np.array([prbox_opt_[0, 0], prbox_opt_[0, 1]]) - np.array(
                        [prbox_opt_[1, 0], prbox_opt_[1, 1]])) * \
                            np.linalg.norm(np.array([prbox_opt_[1, 0], prbox_opt_[1, 1]]) - np.array(
                                [prbox_opt_[2, 0], prbox_opt_[2, 1]]))
                    area_ratio = A_new / A1

                    if area_ratio > 0.1 and area_ratio < 2.5:  # 1.7
                        prbox_opt = prbox_opt_
                    else:
                        print('Bbox optimization has made too large difference.')

            displacement = np.mean(prbox, axis=0) - np.array([mask.shape[0] / 2, mask.shape[1] / 2])
            prbox = (prbox - np.mean(prbox, axis=0) + displacement) / f_ + np.array([pos[1].item(), pos[0].item()])

            if self.params.segm_scale_estimation:

                # use pixels_ratio to determine if new scale should be estimated or not
                mask_pixels_ = np.max(cnt_area)
                pixels_ratio = abs(np.mean(self.mask_pixels) - mask_pixels_) / np.mean(self.mask_pixels)
                if self.uncert_score < self.params.uncertainty_segm_scale_thr:
                    if pixels_ratio < self.params.segm_pixels_ratio:  # 0.6:
                        self.mask_pixels = np.append(self.mask_pixels, mask_pixels_)
                        if self.mask_pixels.size > self.params.mask_pixels_budget_sz:
                            self.mask_pixels = np.delete(self.mask_pixels, 0)

                        new_aabb = self.poly_to_aabbox(prbox[:, 0], prbox[:, 1])
                        new_target_scale = (math.sqrt(new_aabb[2] * new_aabb[3]) * self.params.search_area_scale) / \
                                           self.img_sample_sz[0]
                        rel_scale_ch = (abs(new_target_scale - self.target_scale) / self.target_scale).item()
                        if new_target_scale > self.params.segm_min_scale and rel_scale_ch < self.params.max_rel_scale_ch_thr:
                            self.target_scale = max(self.target_scale * self.params.min_scale_change_factor,
                                                    min(self.target_scale * self.params.max_scale_change_factor,
                                                        new_target_scale))

            if not self.params.segm_scale_estimation or pixels_ratio < self.params.consider_segm_pixels_ratio:
                self.pos[0] = np.mean(prbox[:, 1])
                self.pos[1] = np.mean(prbox[:, 0])

            if not self.params.segm_scale_estimation or pixels_ratio < self.params.segm_pixels_ratio:
                if prbox_opt.size > 0:
                    displacement_opt = np.mean(prbox_opt, axis=0) - np.array([mask.shape[0] / 2, mask.shape[1] / 2])
                    prbox = (prbox_opt - np.mean(prbox_opt, axis=0) + displacement_opt) / f_ + np.array(
                        [pos[1].item(), pos[0].item()])

                if self.rotated_bbox:
                    pred_region = [prbox[0, 0], prbox[0, 1], prbox[1, 0], prbox[1, 1], prbox[2, 0], prbox[2, 1],
                                   prbox[3, 0], prbox[3, 1]]
                else:
                    pred_region = [np.min(prbox[:, 0]) + 1, np.min(prbox[:, 1]) + 1,
                                   np.max(prbox[:, 0]) - np.min(prbox[:, 0]) + 1,
                                   np.max(prbox[:, 1]) - np.min(prbox[:, 1]) + 1]

                return pred_region  # list[8]

        
        """

        return self.vos_mask_result

    def track(self, image, bbox):

        self.pos = torch.Tensor([bbox[1] + bbox[3] / 2, bbox[0] + bbox[2] / 2])
        self.target_sz = torch.Tensor([bbox[3], bbox[2]])

        pred_segm_region = self.segment_target(image, self.pos, self.target_sz)

        return pred_segm_region
