# -*- coding: utf-8 -*
import copy
import itertools
import logging
import math
import os
import os.path as osp
from collections import OrderedDict
from multiprocessing import Process, Queue
from os.path import join

import cv2
import numpy as np
from tqdm import tqdm
from votTester import vot_benchmark
from votTester.pysot.bin.eval import vot_eval

def read_image(image_file: str):
    return cv2.cvtColor(cv2.imread(image_file), cv2.COLOR_BGR2RGB)

def ensure_dir(dir_path: str):
    r"""
    Ensure the existence of path (i.e. mkdir -p)
    Arguments
    ---------
    dir_path: str
        path to be ensured
    """
    if osp.exists(dir_path):
        return
    else:
        os.makedirs(dir_path)

"""
class vot_eval: __init__(self, dataset, dataset_dir, tracker_result_dir, num = 8 ):

"""


class VOTTester:
    r"""
    Tester to test the vot dataset, the result is saved as follows
    exp_dir/logs/$dataset_name$/$tracker_name/
                                    |-baseline/$video_name$/ folder of result files
                                    |-eval_result.csv evaluation result file

    Hyper-parameters
    ----------------
    device_num: int
        number of gpu for test
    data_root: dict
        vot dataset root directory. dict(dataset_name: path_to_root)
    dataset_names: str
        daataset name (VOT2018|VOT2019)
    """

    def __init__(self, tracker, test_name, dataset_names = ['vot2018'], hard_test= 0 , *args, **kwargs):

        self.speed = -1
        self.result_root_dir = '/home/v-xfei/Public/project/pytracking/results/'
        self.data_root = {
            'vot2018':'/home/v-xfei/Public/dataset/vot_tools/vot18/sequences/' ,
            'vot2019':'/home/v-xfei/Public/dataset/vot_tools/vot19/sequences/',
            'vot2016':'/home/v-xfei/Public/dataset/vot_tools/vot16/sequences/'
        }
        self.test_tracker = tracker
        self.dataset_names = dataset_names
        self.test_name = test_name
        self.pysot_eval = vot_eval(self.dataset_names[0], self.data_root[self.dataset_names[0]], os.path.join(self.result_root_dir,self.dataset_names[0]), num=8)
        self.hard_test_flag = hard_test

    def test(self, ):

        for dataset_name in self.dataset_names:
            self.dataset_name = dataset_name
            # self.tracker_dir = os.path.join(self._cfg.auto.log_dir, self._hyper_params["dataset_name"])
            self.save_root_dir = os.path.join(self.result_root_dir,self.dataset_name, self.test_name, "baseline")
            ensure_dir(self.save_root_dir)
            # track videos
            self.run_tracker()
            # evaluation
            save_root_dir = os.path.join(self.result_root_dir, self.dataset_name)
            vot_root = self.data_root[dataset_name]
            vot_eval_show = vot_eval(dataset_name, vot_root, save_root_dir, num=8)
            vot_eval_show.eval(self.test_name)

    def run_tracker(self):
        """
        Run self.pipeline on VOT
        """
        vot_root = self.data_root[self.dataset_name]
        # setup dataset
        dataset = vot_benchmark.load_dataset(vot_root, self.dataset_name)
        self.dataset = dataset
        keys = list(dataset.keys())
        keys.sort()
        if self.hard_test_flag == 1:
            if '19' in self.dataset_name:
                keys = ['agility',
                        'conduction1',
                        'book',
                        'dribble',
                        'flamingo1',
                        'frisbee',
                        'hand2',
                        'handball2',
                        'rabit',
                        'road',
                        ]
        elif self.hard_test_flag == 2:
            if '19' in self.dataset_name:
                keys = ['agility',
                        'dribble',
                        'frisbee',
                        'hand2',
                        'handball2',
                        'rabbit',
                        ]
        elif self.hard_test_flag == 3:
            if '19' in self.dataset_name:
                keys = [
                        'dribble',
                      #  'frisbee',
                        'hand2',
                        ]
      #  keys = []
     #   print('???????????????????????????????????????????????')
        nr_records = len(keys)
        pbar = tqdm(total=nr_records)
        mean_speed = -1
        total_lost = 0
        speed_list = []
        result_queue = Queue(500)
        speed_queue = Queue(500)
        # set worker
        self.worker(keys, result_queue, speed_queue)
        for i in range(nr_records):
            t = result_queue.get()
            s = speed_queue.get()
            total_lost += t
            speed_list.append(s)
            pbar.update(1)

        # print result
        mean_speed = float(np.mean(speed_list))
        print('Total Lost: {:d}'.format(total_lost))
        print('Mean Speed: {:.2f} FPS'.format(mean_speed))
        self.speed= mean_speed

    def worker(self, records, result_queue=None, speed_queue=None):
        r"""
        Worker to run tracker on records

        Arguments
        ---------
        records:
            specific records, can be a subset of whole sequence
        dev: torch.device object
            target device
        result_queue:
            queue for result collecting
        speed_queue:
            queue for fps measurement collecting
        """
        #tracker = copy.deepcopy(self._pipeline)
        tracker = self.test_tracker
        sucess = []
        for v_id, video in enumerate(records):
            lost, speed, sucess_pescent= self.track_single_video(tracker, video, v_id=v_id)
            if result_queue is not None:
                result_queue.put_nowait(lost)
            if speed_queue is not None:
                speed_queue.put_nowait(speed)
            sucess.append(sucess_pescent)

        def mean(ls):
            sum=0
            for item in ls:
                sum+=item
            return sum/len(ls)

        print('average 60 success percent: ' + str(mean(sucess)))


    def evaluation(self):
        for dataset_name in self.dataset_names:
            save_root_dir = os.path.join(self.result_root_dir,self.dataset_name)
            vot_root = self.data_root[self.dataset_name]
            vot_eval_show = vot_eval(dataset_name, vot_root, save_root_dir, num = 8)
            vot_eval_show.eval(self.test_name)

    def track_single_video(self, tracker, video, v_id=0):
        regions = []
        video = self.dataset[video]
        image_files_video, gt_video = video['image_files'], video['gt']
        start_frame_video, end_frame_video, lost_times, toc = 0, len(image_files_video), 0, 0

        anchor_num = round(((end_frame_video + 1) / 50)) - 1
        run_times = anchor_num + 2
        for run_index in range(run_times):
            sucess_pescent = None

            if run_index ==0:
                start_frame = 0
                end_frame = end_frame_video
                image_files = image_files_video
                gt = gt_video
            elif run_index == run_times-1:
                start_frame = 0
                end_frame = end_frame_video
                image_files = image_files_video[::-1]
                gt = gt_video[::-1]
            else:
                anchor = 50 * (run_index)
                anchor_remain = end_frame_video - anchor
                if anchor_remain >= anchor:
                    start_frame = 0
                    end_frame = anchor_remain
                    image_files = image_files_video[(anchor-1):end_frame_video:1]
                    gt = gt_video[(anchor-1):end_frame_video:1]
                else:
                    start_frame = 0
                    end_frame = anchor
                    image_files = image_files_video[0:anchor]
                    image_files.reverse()
                    gt = gt_video[0:anchor][::-1]
            for f, image_file in enumerate(tqdm(image_files)):
                ###############################注意 读取图片 有无更换通道位置
                im_rgb = cv2.cvtColor(cv2.imread(image_file), cv2.COLOR_BGR2RGB)
                im_bgr = cv2.imread(image_file)
                tic = cv2.getTickCount()
                if f == start_frame:  # init
                    cx, cy, w, h = vot_benchmark.get_axis_aligned_bbox(gt[f])
                    location = vot_benchmark.cxy_wh_2_rect((cx, cy), (w, h))  # x0, y0 , w, h  > left top

                    list_8 = gt[f].tolist()
                    list_4 = location.tolist()
                    tracker.init_votest(im_rgb, im_bgr, list_8, list_4, video['name'])
                    regions.append(1 if 'VOT' in self.dataset_name else gt[f])
                    gt_polygon = None
                    pred_polygon = None
                elif f > start_frame:  # tracking

                    report_loc = tracker.update_votest(im_rgb, im_bgr)

                    if len(report_loc) == 4:
                        gt_polygon = (gt[f][0], gt[f][1], gt[f][2], gt[f][3], gt[f][4],
                                      gt[f][5], gt[f][6], gt[f][7])
                        pred_polygon_outpoly = (report_loc[0], report_loc[1],
                                                report_loc[0] + report_loc[2], report_loc[1],
                                                report_loc[0] + report_loc[2], report_loc[1] + report_loc[3],
                                                report_loc[0], report_loc[1] + report_loc[3])

                    elif len(report_loc) == 8:
                        gt_polygon = (gt[f][0], gt[f][1], gt[f][2], gt[f][3], gt[f][4],
                                      gt[f][5], gt[f][6], gt[f][7])
                        pred_polygon_outpoly = report_loc
                    else:
                        print('error update format')

                    b_overlap = vot_benchmark.vot_overlap(
                        gt_polygon, pred_polygon_outpoly, (im_rgb.shape[1], im_rgb.shape[0]))
                    if b_overlap > 0.1:
                        regions.append(pred_polygon_outpoly)  # pred_polygon
                    else:  # lost
                        regions.append(2)
                        lost_times += 1
                        # start_frame = f + 5  # skip 5 frames
                        start_frame = f + 100000  # skip 5 frames
                        sucess_pescent = f / (end_frame - 1)

                else:  # skip
                    regions.append(0)
                toc += cv2.getTickCount() - tic

            if sucess_pescent is None: sucess_pescent=1.0

            toc /= cv2.getTickFrequency()

            # save result
            result_dir = join(self.save_root_dir, video['name'])
            ensure_dir(result_dir)
            result_path = join(result_dir, '{:s}_001.txt'.format(video['name']))
            with open(result_path, "w") as fin:
                 for x in regions:
                     fin.write("{:d}\n".format(x)) if isinstance(x, int) else \
                     fin.write(','.join([vot_benchmark.vot_float2str("%.4f", i) for i in x]) + '\n')

            print('this short video situation:')
            print(video['name'] + ": frame: "+ image_files[0][-8:-4:1] + 'to'+ image_files[-1][-8:-4:1])
            print('({:d}) Video: {:12s} Time: {:02.1f}s Speed: {:3.1f}fps Lost: {:d} '
                  .format(v_id, video['name'] + ' anchor: ' + str(run_index+1), toc, f / toc, lost_times))
            print('success percent: ' + str(sucess_pescent))

        return lost_times, f / toc, sucess_pescent

